%% A k-space filtering technique for vibration sonoelastography
%
% Master thesis for Mikkel Henninen June 2016
%
% authors: Alfonso Rodriguez-Molares <alfonso.r.molares@ntnu.no>
%          Mikkel Henninen <mikkeloy@stud.ntnu.no>

clear all;
close all;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% File loop
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
for filenumber = 6:6 %3.5, 2.5, 1, 0.5, hard,soft

    tic

    clearvars -except filter_iter filenumber
    saving = 0;
    add_noise=0;
    use_fitting=1;      

    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% Loading/downloading data
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    url='http://folk.ntnu.no/alfonsom/elastography_data/numerical/';

    %Change here for you local path
    folder='Data\numerical\';
    savefolder = 'C:\Users\Min\Documents\MATLAB\Master\Figs\Final results\Numerical_alfonso\';


    switch (filenumber)
        case 1
    filename='COMSOL_hardcyst_damped_3500.txt'; acoustic_frequency = 3500; soft=false; save_name = 'hard_cyst_3500_';
        case 2
    filename='COMSOL_hardcyst_damped_2500.txt'; acoustic_frequency = 2500; soft=false; save_name = 'hard_cyst_2500_';
        case 3
    filename='COMSOL_hardcyst_damped_1000.txt'; acoustic_frequency = 1000; soft=false; save_name = 'hard_cyst_1000_';
        case 4
    filename='COMSOL_hardcyst_damped_500.txt'; acoustic_frequency = 500; soft=false; save_name = 'hard_cyst_500_';
        case 5
    filename='COMSOL_softcyst_damped_3500.txt'; acoustic_frequency = 3500; soft=true; save_name = 'soft_cyst_3500_';
        case 6
    filename='COMSOL_softcyst_damped_2500.txt'; acoustic_frequency = 2500; soft=true; save_name = 'soft_cyst_2500_';
        case 7
    filename='COMSOL_softcyst_damped_1000.txt'; acoustic_frequency = 1000; soft=true; save_name = 'soft_cyst_1000_';
        case 8
    filename='COMSOL_softcyst_damped_500.txt'; acoustic_frequency = 500; soft=true; save_name = 'soft_cyst_500_';
    end
    % we check if the file is on disk; if not we download it from the webpage
    a=dir([folder filename]);
    if not(numel(a))
        disp('Downloading data...');
        urlwrite([url filename],[folder filename]);
    end

    [x,z,u,ui,v,vi,w] = textread([folder filename],'%f %f %f %fi %f %fi %f','commentstyle','matlab');
    u=u+1i*ui;
    v=v+1i*vi;
    z=-z;
    
    % check if the case has been calculated
    procesdatafilename=[save_name 'workspace_after_fitting_and_kspace.mat'];
    if numel(dir(procesdatafilename))
        load(procesdatafilename);
    else

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% Interpolation on a regular grid
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        cs_min=1;                              % minimum shear wave speed we will visualize [m/s]
        cs_max=20;                             % minimum shear wave speed we will visualize [m/s]
        lambda_max=cs_max/acoustic_frequency;   
        lambda_min=cs_min/acoustic_frequency;
        oversampling=2;                  % oversampling factor
        Fs=oversampling/lambda_min;      % sampling spatial frequency [Hz]
        ds=1/Fs;                         % spatial step [m]

        % spatial sampling
        x_axis=(-20e-3:ds:20e-3).';     % x vector [m]
        z_axis=(0:ds:40e-3).';             % z vector [m]
        [X Z]=meshgrid(x_axis,z_axis);
        U=griddata(x,z,u,X,Z,'linear');

        % repare NaN values inserted by interpolation
        repair_mask=isnan(U);
        U(repair_mask)=griddata(x,z,u,X(repair_mask),X(repair_mask),'nearest');

        % ground truth
        cyst_radius=7.5e-3;
        cx=0; cz=20e-3;
        mask=sqrt((X-cx).^2+(Z-cz).^2)<cyst_radius;
        cs1=5;
        cs2=10;
        if soft
            ground_truth=cs2*~mask+cs1*mask;
        else
            ground_truth=cs1*~mask+cs2*mask;
        end

        % figure;
        % imagesc(x_axis*1e3,z_axis*1e3,real(U/max(abs(U(:))))); shading flat; axis equal tight; colorbar; title('Normalized axial velocity'); colorbar; hold on;
        % plot(cx*1e3+cyst_radius*1e3*sin(linspace(0,2*pi,180)),cz*1e3+cyst_radius*1e3*cos(linspace(0,2*pi,180)),'r--')

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% Extrapolating other frames
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        n_frames=50;                             % number of frames % MIKKEL: This number can be reduced to reduce computation time  
        for n=1:n_frames
            TDI(:,:,n)=real(U.*exp(j*2*pi*n/n_frames));
        end
        [Nz Nx Nf]=size(TDI);

        % % showing original signal
        % max_val=max(abs(U(:)));
        % figure;
        % colormap jet;
        % for n=1:size(TDI,3)
        %     imagesc(x_axis*1e3,z_axis*1e3,TDI(:,:,n)); shading flat; axis equal tight; shading flat;
        %     caxis([-max_val max_val]);
        %     colorbar;
        %     title(sprintf('Frame %d',n));
        %     pause(0.1)
        % end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% adding high frequency noise
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        if (add_noise)

            TDI=TDI+0.1*max_val.*random('normal',0,1,size(TDI));

%             % showing signal + noise
%             figure;
%             colormap jet;
%             for n=1:size(TDI,3)
%                 pcolor(x_axis*1e3,z_axis*1e3,TDI(:,:,n)); axis equal tight; shading flat;
%                 caxis([-max_val max_val]);
%                 colorbar;
%                 title(sprintf('Original, frame %d',n));
%                 pause(0.1)
%             end
        end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% Temporal filtering 
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        % Removes the presence of noise and other frequencies

        if (use_fitting)

            ttdi=reshape(permute(TDI,[3 1 2]),[ size(TDI,3) size(TDI,1)*size(TDI,2)]);

            % sine fit
            ftdi=zeros(size(ttdi));
            amplitude=zeros(size(ttdi,1),1);
            phase=zeros(size(ttdi,1),1);
            t=linspace(0,1/acoustic_frequency,size(ttdi,1));
            disp('Starting the parallel processing. This will take a while...');
            parfor n=1:size(ttdi,2)
                options = fitoptions('sin1');
                options.Lower = [-Inf 2*pi*acoustic_frequency*0.99 -Inf];
                options.Upper = [+Inf 2*pi*acoustic_frequency*1.01 +Inf];
                model=fit(t.',ttdi(:,n),'sin1',options);
                amplitude(n)=model.a1;
                phase(n)=model.c1;
            end

            AMP=reshape(amplitude,[size(TDI,1) size(TDI,2)]);
            PHA=reshape(phase,[size(TDI,1) size(TDI,2)]);

            % spatial smoothing of phase and amplitude
            Ns=3;

            if 0
                % mean filter
                sAvg = ones(Ns)./Ns^2;                  
                AMP = filter2(sAvg,AMP);  
                PHA = filter2(sAvg,PHA);  
            else
                % median filter
                AMP = medfilt2(AMP,[Ns Ns]);  
                PHA = medfilt2(PHA,[Ns Ns]);  
            end

        %     figure;
        %     subplot(1,2,1);
        %     pcolor(x_axis*1e3,z_axis*1e3,AMP); axis equal tight; shading flat;
        %     colorbar;
        %     title('Amplitude');
        %     subplot(1,2,2);
        %     pcolor(x_axis*1e3,z_axis*1e3,PHA); axis equal tight; shading flat;
        %     colorbar;
        %     title('Phase');

            % generating the smoothed signal
            PTDI=zeros(size(TDI));
            for n=1:size(TDI,3)
                %PTDI(:,:,n)=AMP.*sin(2*pi*acoustic_frequency*t(n)+PHA);
                PTDI(:,:,n)=sin(2*pi*acoustic_frequency*t(n)+PHA);      
            end

        %     % showing spatial averaged signal
        %     max_val=max(abs(TDI(:)));
        %     figure;
        %     colormap jet;
        %     for n=1:size(TDI,3)
        %         subplot(1,2,1);
        %         pcolor(x_axis*1e3,z_axis*1e3,TDI(:,:,n)); axis equal tight; shading flat;
        %         caxis([-max_val max_val]);
        %         colorbar;
        %         title(sprintf('Original, frame %d',n));
        % 
        %         subplot(1,2,2);
        %         pcolor(x_axis*1e3,z_axis*1e3,PTDI(:,:,n)); axis equal tight; shading flat;
        %         caxis([-1 1]);
        %         colorbar;
        %         title(sprintf('Filtered, frame %d',n));
        % 
        %         pause(0.1)
        %     end
        end

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% k-space transform
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        % MIKEL: You were using an old version of the k-space transform. Not
        % that there was something wrong with it, but I think I have made some
        % small improvements in the function "kspace" below.

        %% compute the k-space transform
        disp('Computing k-space. This may take a while');
        [kfield,kz,kx]=kspace(PTDI,ds,0.25,4,false);  
        % showing kspace
        theta=linspace(0,2*pi,180);
        ks1=2*pi*acoustic_frequency/cs1;
        ks2=2*pi*acoustic_frequency/cs2;

        figure;
        pcolor(kx*1e-3,kz*1e-3,sum(abs(kfield),3)); axis equal tight; hold on; shading flat;
        xlabel('k_x [rad/mm]');ylabel('k_z [rad/mm]');
        plot(ks1*cos(theta)*1e-3,ks1*sin(theta)*1e-3,'y--'); 
        plot(ks2*cos(theta)*1e-3,ks2*sin(theta)*1e-3,'r--'); 
        xlim([-2*ks1 2*ks1]/1e3);
        ylim([-2*ks1 2*ks1]/1e3);
        title('k-space');
        legend('',sprintf('cs=%0.1f m/s',cs1),sprintf('cs=%0.1f m/s',cs2));
        drawnow;

        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% spectrum estimation
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


        alpha=linspace(0,2*pi,180);
        c_axis=linspace(cs_max,1,200);
        k_axis=2*pi*acoustic_frequency./c_axis;
        kx_axis=k_axis.'*sin(alpha); kz_axis=k_axis.'*cos(alpha);

        spectrum=zeros(size(c_axis)).';
        for n=1:size(kfield,3)
            kf=interp2(kx,kz,kfield(:,:,n),kx_axis,kz_axis,'linear');
            spectrum=spectrum+(sum(abs(kf),2)); 
        end

        figure;
        plot(k_axis,spectrum/max(spectrum),'k-'); hold on; grid on;
        plot([2*pi*acoustic_frequency/cs2 2*pi*acoustic_frequency/cs2],[0 1],'b--');
        plot([2*pi*acoustic_frequency/cs1 2*pi*acoustic_frequency/cs1],[0 1],'r--');
        xlabel('k [rad/m]');
        legend('cyst + background','cyst','background');
        title(sprintf('k-space spectrum %0.0f Hz',acoustic_frequency));

        % remove NaNs
        spectrum(isnan(spectrum))=0;
    
        % save workspace for future re-runs
        save(procesdatafilename);
    end
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %% feature extraction
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
     
        datafilename=[save_name];
                
        cs_vector=[];
        filter=[];
            %% peak detection
            %
            

            % first we smothen the spectrum a bit to remove noisy peaks
            smoothing_window_size=15;
            f_spectrum=filtfilt(hanning(smoothing_window_size)/sum(hanning(smoothing_window_size)),1,spectrum);

            % we detect the peaks
            [pks,locs,widths,proms] = findpeaks(f_spectrum,k_axis);

            % we sort them by prominence (i.e. importance)
            [dummy indexes]=sort(proms,'descend');
            pks=pks(indexes);
            locs=locs(indexes);
            widths=widths(indexes);
            proms=proms(indexes);
            cs_vector=locs;

            % design ring filter automatically
            for m=1:length(locs)
                filter(m,:)=(k_axis>locs(m)-widths(m))&(k_axis<locs(m)+widths(m));
            end

                % show spectrum, peaks & filters
                figure;
                plot(c_axis,spectrum./max(spectrum),'k-'); hold on; grid on;
                plot(c_axis,f_spectrum./max(spectrum),'b--'); 
                plot(2*pi*acoustic_frequency./locs,pks./max(spectrum),'r^'); 
                for m=1:length(locs)
                    plot(2*pi*acoustic_frequency./[locs(m)-widths(m)/2 locs(m)+widths(m)/2],[pks(m) pks(m)]./max(spectrum),'r-s'); 
                    plot(c_axis,filter(m,:),'--');
                end
                xlabel('Shear wave speed [m/s]');
                ylabel('Normalized energy');



        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %% Filtering
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        components=2;%size(filter,1)
        [KX KZ]=meshgrid(kx,kz);
        K=sqrt(KX.^2+KZ.^2);
        cs=zeros(1,1,components);
        map=zeros(size(TDI,1),size(TDI,2),components);
        for m=1:components
            % filtering
            kfiltered=bsxfun(@times,kfield,interp1(k_axis,filter(m,:),K,'linear',0));

            % inverse fft
            disp('Computing inverse FFT. This may take a while');
            ffield=ifft2(ifftshift(kfiltered));
            ffield=real(ifft2(ifftshift(kfiltered)));
            ffield=ffield(1:Nz,1:Nx,:,:);

            % storing map and
            cs(1,1,m)=2*pi*acoustic_frequency/cs_vector(m);
            map(:,:,m)=sum(abs(ffield),3);
% 
% % %             % showing space
% % %             figure(106);
% % %             subplot(1,4,1); hold off; 
% % %             imagesc(x_axis,z_axis,ctdi); axis equal tight; hold on; 
% % %             title('Space');
% %             % showing kspace
% %             subplot(1,4,2); hold off;
%             imagesc(kx*1e-3,kz*1e-3,abs(kfield(:,:,i))); axis equal tight; hold on; shading flat;
%             xlabel('k_x [rad/mm]');ylabel('k_z [rad/mm]');
%             xlim([-ks1 ks1]*1.5*1e-3);
%             ylim([-ks1 ks1]*1.5*1e-3);
%             title('k-space');
% %             pause()
% %             % showing filtered kspace
% %             subplot(1,4,3); hold off;
%             imagesc(kx*1e-3,kz*1e-3,abs(kfiltered(:,:,i))); axis equal tight; hold on; shading flat;
%             xlabel('k_x [rad/mm]');ylabel('k_z [rad/mm]');
%             xlim([-3.5 3.5]);
%             ylim([-3.5 3.5]);
%             title('k-space of the second filtered component');   
% %             pause
%             % showing space
%             %subplot(1,4,4); hold off; 
% %             imagesc(x_axis*1e3,z_axis*1e3,ffield(:,:,n)); axis equal tight; colorbar;
% %             title('Filtered Velocity space of the first component');
% %             xlabel('x [mm]')
% %             ylabel('z [mm]')
% %        pause(0.1)
% 
%             figure(121);
%             subplot(1,2,m)
%             imagesc(x_axis*1e3,z_axis*1e3,map(:,:,m)); axis equal tight; hold on; 
%             plot(cx*1e3+cyst_radius*1e3*sin(linspace(0,2*pi,180)),cz*1e3+cyst_radius*1e3*cos(linspace(0,2*pi,180)),'r--')
%             xlabel('x [mm]');
%             ylabel('z [mm]');
%             title(sprintf('Filtered region for %0.1f m/s',cs(m)));
% %             if saving && m==2
% %               saveas(gcf,[savefolder sprintf('%smethod_%d_filter_%d_double_test',save_name,extraction_method,m)], 'jpg');
% %             end
        end
        
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        %%  Global map
        %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
        
        % computing global map as a weigthed mean
        csm=sum(bsxfun(@times,map,cs),3)./sum(map,3);
    
        % smoothing a bit with a media filter        
        csmm=medfilt2(csm,[5 5]);

        figure(102);
        imagesc(x_axis*1e3,z_axis*1e3,csmm); axis equal tight; shading flat; colormap jet; hold on;
        plot(cx*1e3+cyst_radius*1e3*sin(linspace(0,2*pi,180)),cz*1e3+cyst_radius*1e3*cos(linspace(0,2*pi,180)),'k--')
        caxis([5 10]);
        c=colorbar; xlabel(c,'c_s [m/s]');
        xlabel('x [mm]');
        ylabel('z [mm]');
        title('Cumulative velocity map [m/s]');
        if saving
        saveas(gcf,[savefolder sprintf('%smethod_%d_elasticity_map',save_name,extraction_method)], 'jpg');
        end
    end          
end

  